﻿using System;

namespace Niche.Common
{
    /// <summary>
    /// Generic wrapper for storing scoped values
    /// </summary>
    /// <typeparam name="T">Type of value to scope.</typeparam>
    public class Scoped<T>
    {
        /// <summary>
        /// Gets the current scoped value
        /// </summary>
        public T Value
        {
            get
            {
                return mValue;
            }

            private set
            {
                mValue = value;
                OnValueChanged(mValue);
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Scoped{T}"/> class.
        /// </summary>
        /// <param name="defaultValue">The default value.</param>
        public Scoped(T defaultValue)
        {
            Value = defaultValue;
        }

        /// <summary>
        /// Specify a new value for the scope, returning an action that 
        /// restores the pre-existing value.
        /// </summary>
        /// <param name="scopedValue">
        /// Value to set for the duration of the scope.
        /// </param>
        /// <returns>Disposable action </returns>
        public IDisposable With(T scopedValue)
        {
            T currentValue = Value;
            Value = scopedValue;
            return new DisposableAction(
                    () =>
                    {
                        Value = currentValue;
                    });
        }

        /// <summary>
        /// Alternative syntax to specify a new value for the scope, returning 
        /// an action that restores the pre-existing value.
        /// </summary>
        /// <param name="scopedValue">
        /// Value to set for the duration of the scope.
        /// </param>
        /// <returns>Disposable action </returns>
        public IDisposable As(T scopedValue)
        {
            return With(scopedValue);
        }

        /// <summary>
        /// Event triggered when our contained value changes
        /// </summary>
        public event EventHandler<ValueChangedEventArgs<T>> ValueChanged
        {
            add
            {
                mValueChanged += value;
            }

            remove
            {
                mValueChanged -= value;
            }
        }

        /// <summary>
        /// Trigger method for the MyEvent event.
        /// </summary>
        /// <param name="value">Value that has been set.</param>
        protected void OnValueChanged(T value)
        {
            var handler = mValueChanged;
            if (handler != null)
            {
                var args = new ValueChangedEventArgs<T>(value);
                handler(this, args);
            }
        }

        /// <summary>
        /// Storage for the MyEvent event.
        /// </summary>
        private EventHandler<ValueChangedEventArgs<T>> mValueChanged;

        /// <summary>
        /// Storage for the Value property
        /// </summary>
        private T mValue;
    }
}
